@@ -206,10 +206,9 @@ private:
     bool Error = false;              ///< Could not allocate.
 
     explicit LiveReg(Register VirtReg) : VirtReg(VirtReg) {}
+    explicit LiveReg() {}
 
-    unsigned getSparseSetIndex() const {
-      return Register::virtReg2Index(VirtReg);
-    }
+    unsigned getSparseSetIndex() const { return VirtReg.virtRegIndex(); }
   };
 
   using LiveRegMap = SparseSet<LiveReg, identity<unsigned>, uint16_t>;
@@ -218,101 +217,12 @@ private:
   LiveRegMap LiveVirtRegs;
 
   /// Stores assigned virtual registers present in the bundle MI.
-  DenseMap<Register, MCPhysReg> BundleVirtRegsMap;
+  DenseMap<Register, LiveReg> BundleVirtRegsMap;
 
-  DenseMap<unsigned, SmallVector<MachineOperand *, 2>> LiveDbgValueMap;
+  DenseMap<Register, SmallVector<MachineOperand *, 2>> LiveDbgValueMap;
   /// List of DBG_VALUE that we encountered without the vreg being assigned
   /// because they were placed after the last use of the vreg.
-  DenseMap<unsigned, SmallVector<MachineInstr *, 1>> DanglingDbgValues;
-
-  /// Maps unspilled virtual registers to all DbgDefs which originally used
-  /// the virtual register as their referrer. Used to allow updating DbgDefs
-  /// which were visited before the spill occurs.
-  DenseMap<unsigned, SmallVector<MachineInstr *, 2>> VReg2NonStackDbgDefs;
-  /// Subset of VReg2NonStackDbgDefs. Maps unspilled virtual registers to
-  /// DbgDefs which occured after the last non-debug use of the virtual
-  /// register. Allows for some such "dangling" defs to be salvaged after the
-  /// block is allocated.
-  DenseMap<unsigned, SmallVector<MachineInstr *, 2>> VReg2DanglingDbgDefs;
-  /// Essentially a typed pair of MachineInstr, representing which existing
-  /// or inserted DbgKills exist for a given Lifetime.
-  class DbgKills final {
-    /// The current expected "live" DbgKill. This is either:
-    /// * Empty (if we have not seen any existing DbgKill or non-stack DbgDef
-    /// which would require us to insert one).
-    /// * An existing DbgKill, if we encounter it before the DbgDef.
-    /// * An inserted DbgKill, if we encounter the DbgDef before any existing
-    /// DbgKill, or if there is no such existing DbgKill.
-    ///
-    /// Read as `std::optional<MachineInstr>`.
-    MachineInstr *Live = nullptr;
-    /// The existing DbgKill, if one exists and we have also inserted a
-    /// DbgKill for a non-stack DbgDef. This is retained until we either:
-    /// * Spill the DbgDef, at which point we can remove our inserted DbgKill
-    /// and use the existing one.
-    /// * Finish allocating the function, at which point we can remove this
-    /// "backup" existing DbgDef instruction.
-    ///
-    /// Read as `std::optional<MachineInstr>`
-    MachineInstr *Dead = nullptr;
-    /// Flag to represent the case where we want to eliminate all future
-    /// DbgKills we encounter, as we have removed the corresponding DbgDef
-    /// (i.e. we have "dropped" this debug info).
-    bool RemoveAll = false;
-
-  public:
-    /// Update to reflect an inserted DbgKill, marking an existing DbgKill to
-    /// be removed if it exists and we have already seen it.
-    void handleNewDbgKill(MachineInstr &DbgKill) {
-      assert(!RemoveAll);
-      if (Live)
-        Dead = Live;
-      Live = &DbgKill;
-    }
-    /// Update to reflect an existing DbgKill, either marking it to be
-    /// removed (if we have already inserted a corresponding DbgKill) or
-    /// recording it in case we need to remove it when we later process a
-    /// corresponding DbgDef.
-    void handleExistingDbgKill(MachineInstr &DbgKill) {
-      if (RemoveAll) {
-        DbgKill.removeFromParent();
-        return;
-      }
-      assert(!Dead);
-      if (Live)
-        Dead = &DbgKill;
-      else
-        Live = &DbgKill;
-    }
-    /// Handle when the virtual register referrer of the corresponding DbgDef
-    /// is spilled by reverting back to an existing DbgKill if it exists.
-    void handleSpill() {
-      if (!Dead)
-        return;
-      Live->removeFromParent();
-      Live = Dead;
-      Dead = nullptr;
-    }
-    /// Update to reflect that the corresponding DbgDef has been eliminated.
-    void removeAll() {
-      RemoveAll = true;
-      if (Live)
-        Live->removeFromParent();
-      if (Dead)
-        Dead->removeFromParent();
-      Live = Dead = nullptr;
-    }
-    /// Update to reflect that allocation is complete, removing Dead if it
-    /// exists.
-    void removeDead() {
-      if (Dead)
-        Dead->removeFromParent();
-      Dead = nullptr;
-    }
-  };
-  /// Maps a lifetime (for which there can be at most one DbgDef, and up to
-  /// two DbgKills during allocation) to the DbgKills object which tracks it.
-  DenseMap<DILifetime *, DbgKills> Lifetime2DbgKills;
+  DenseMap<Register, SmallVector<MachineInstr *, 1>> DanglingDbgValues;
 
   /// Has a bit set for every virtual register for which it was determined
   /// that it is alive across blocks.
@@ -366,7 +276,7 @@ private:
   InstrPosIndexes PosIndexes;
 
   void setPhysRegState(MCRegister PhysReg, unsigned NewState);
-  bool isPhysRegFree(MCPhysReg PhysReg) const;
+  bool isPhysRegFree(MCRegister PhysReg) const;
 
   /// Mark a physreg as used in this instruction.
   void markRegUsedInInstr(MCPhysReg PhysReg) {
@@ -375,14 +285,14 @@ private:
   }
 
   // Check if physreg is clobbered by instruction's regmask(s).
-  bool isClobberedByRegMasks(MCPhysReg PhysReg) const {
+  bool isClobberedByRegMasks(MCRegister PhysReg) const {
     return llvm::any_of(RegMasks, [PhysReg](const uint32_t *Mask) {
       return MachineOperand::clobbersPhysReg(Mask, PhysReg);
     });
   }
 
   /// Check if a physreg or any of its aliases are used in this instruction.
-  bool isRegUsedInInstr(MCPhysReg PhysReg, bool LookAtPhysRegUses) const {
+  bool isRegUsedInInstr(MCRegister PhysReg, bool LookAtPhysRegUses) const {
     if (LookAtPhysRegUses && isClobberedByRegMasks(PhysReg))
       return true;
     for (MCRegUnit Unit : TRI->regunits(PhysReg))
@@ -393,7 +303,7 @@ private:
 
   /// Mark physical register as being used in a register use operand.
   /// This is only used by the special livethrough handling code.
-  void markPhysRegUsedInInstr(MCPhysReg PhysReg) {
+  void markPhysRegUsedInInstr(MCRegister PhysReg) {
     for (MCRegUnit Unit : TRI->regunits(PhysReg)) {
       assert(UsedInInstr[Unit] <= InstrGen && "non-phys use before phys use?");
       UsedInInstr[Unit] = InstrGen;
@@ -401,7 +311,7 @@ private:
   }
 
   /// Remove mark of physical register being used in the instruction.
-  void unmarkRegUsedInInstr(MCPhysReg PhysReg) {
+  void unmarkRegUsedInInstr(MCRegister PhysReg) {
     for (MCRegUnit Unit : TRI->regunits(PhysReg))
       UsedInInstr[Unit] = 0;
   }
@@ -428,37 +338,29 @@ private:
 
   void allocateInstruction(MachineInstr &MI);
   void handleDebugValue(MachineInstr &MI);
-  void killDebugDefWithinBlock(MachineInstr &MI);
-  void handleDebugDefKill(MachineInstr &MI);
   void handleBundle(MachineInstr &MI);
 
-  bool usePhysReg(MachineInstr &MI, MCPhysReg PhysReg);
-  bool definePhysReg(MachineInstr &MI, MCPhysReg PhysReg);
-  bool displacePhysReg(MachineInstr &MI, MCPhysReg PhysReg);
-  void freePhysReg(MCPhysReg PhysReg);
+  bool usePhysReg(MachineInstr &MI, MCRegister PhysReg);
+  bool definePhysReg(MachineInstr &MI, MCRegister PhysReg);
+  bool displacePhysReg(MachineInstr &MI, MCRegister PhysReg);
+  void freePhysReg(MCRegister PhysReg);
 
   unsigned calcSpillCost(MCPhysReg PhysReg) const;
 
   LiveRegMap::iterator findLiveVirtReg(Register VirtReg) {
-    return LiveVirtRegs.find(Register::virtReg2Index(VirtReg));
+    return LiveVirtRegs.find(VirtReg.virtRegIndex());
   }
 
   LiveRegMap::const_iterator findLiveVirtReg(Register VirtReg) const {
-    return LiveVirtRegs.find(Register::virtReg2Index(VirtReg));
+    return LiveVirtRegs.find(VirtReg.virtRegIndex());
   }
 
-  void assignVirtToPhysReg(MachineInstr &MI, LiveReg &, MCPhysReg PhysReg);
+  void assignVirtToPhysReg(MachineInstr &MI, LiveReg &, MCRegister PhysReg);
   void allocVirtReg(MachineInstr &MI, LiveReg &LR, Register Hint,
                     bool LookAtPhysRegUses = false);
   void allocVirtRegUndef(MachineOperand &MO);
   void assignDanglingDebugValues(MachineInstr &Def, Register VirtReg,
-                                 MCPhysReg Reg);
-  void removeDebugDef(MachineInstr &DbgDef) {
-    Lifetime2DbgKills[DbgDef.getDebugLifetime()].removeAll();
-    DbgDef.removeFromParent();
-  }
-  void assignDanglingDebugDefs(MachineInstr &Def, Register VirtReg,
-                               MCPhysReg Reg);
+                                 MCRegister Reg);
   bool defineLiveThroughVirtReg(MachineInstr &MI, unsigned OpNum,
                                 Register VirtReg);
   bool defineVirtReg(MachineInstr &MI, unsigned OpNum, Register VirtReg,
@@ -473,7 +375,8 @@ private:
                             SmallSet<Register, 2> &PrologLiveIns) const;
 
   void reloadAtBegin(MachineBasicBlock &MBB);
-  bool setPhysReg(MachineInstr &MI, MachineOperand &MO, MCPhysReg PhysReg);
+  bool setPhysReg(MachineInstr &MI, MachineOperand &MO,
+                  const LiveReg &Assignment);
 
   Register traceCopies(Register VirtReg) const;
   Register traceCopyChain(Register Reg) const;
@@ -488,6 +391,8 @@ private:
   bool mayLiveOut(Register VirtReg);
   bool mayLiveIn(Register VirtReg);
 
+  bool mayBeSpillFromInlineAsmBr(const MachineInstr &MI) const;
+
   void dumpState() const;
 };
 
@@ -512,22 +417,19 @@ public:
   }
 
   MachineFunctionProperties getRequiredProperties() const override {
-    return MachineFunctionProperties().set(
-        MachineFunctionProperties::Property::NoPHIs);
+    return MachineFunctionProperties().setNoPHIs();
   }
 
   MachineFunctionProperties getSetProperties() const override {
     if (Impl.ClearVirtRegs) {
-      return MachineFunctionProperties().set(
-          MachineFunctionProperties::Property::NoVRegs);
+      return MachineFunctionProperties().setNoVRegs();
     }
 
     return MachineFunctionProperties();
   }
 
   MachineFunctionProperties getClearedProperties() const override {
-    return MachineFunctionProperties().set(
-        MachineFunctionProperties::Property::IsSSA);
+    return MachineFunctionProperties().setIsSSA();
   }
 };
 
@@ -551,7 +453,7 @@ void RegAllocFastImpl::setPhysRegState(MCRegister PhysReg, unsigned NewState) {
     RegUnitStates[Unit] = NewState;
 }
 
-bool RegAllocFastImpl::isPhysRegFree(MCPhysReg PhysReg) const {
+bool RegAllocFastImpl::isPhysRegFree(MCRegister PhysReg) const {
   for (MCRegUnit Unit : TRI->regunits(PhysReg)) {
     if (RegUnitStates[Unit] != regFree)
       return false;
@@ -572,6 +474,13 @@ int RegAllocFastImpl::getStackSpaceFor(Register VirtReg) {
   const TargetRegisterClass &RC = *MRI->getRegClass(VirtReg);
   unsigned Size = TRI->getSpillSize(RC);
   Align Alignment = TRI->getSpillAlign(RC);
+
+  const MachineFunction &MF = MRI->getMF();
+  auto &ST = MF.getSubtarget();
+  Align CurrentAlign = ST.getFrameLowering()->getStackAlign();
+  if (Alignment > CurrentAlign && !TRI->canRealignStack(MF))
+    Alignment = CurrentAlign;
+
   int FrameIdx = MFI->CreateSpillStackObject(Size, Alignment);
 
   // Assign the slot.
@@ -588,9 +497,24 @@ static bool dominates(InstrPosIndexes &PosIndexes, const MachineInstr &A,
   return IndexA < IndexB;
 }
 
+/// Returns true if \p MI is a spill of a live-in physical register in a block
+/// targeted by an INLINEASM_BR. Such spills must precede reloads of live-in
+/// virtual registers, so that we do not reload from an uninitialized stack
+/// slot.
+bool RegAllocFastImpl::mayBeSpillFromInlineAsmBr(const MachineInstr &MI) const {
+  int FI;
+  auto *MBB = MI.getParent();
+  if (MBB->isInlineAsmBrIndirectTarget() && TII->isStoreToStackSlot(MI, FI) &&
+      MFI->isSpillSlotObjectIndex(FI))
+    for (const auto &Op : MI.operands())
+      if (Op.isReg() && MBB->isLiveIn(Op.getReg()))
+        return true;
+  return false;
+}
+
 /// Returns false if \p VirtReg is known to not live out of the current block.
 bool RegAllocFastImpl::mayLiveOut(Register VirtReg) {
-  if (MayLiveAcrossBlocks.test(Register::virtReg2Index(VirtReg))) {
+  if (MayLiveAcrossBlocks.test(VirtReg.virtRegIndex())) {
     // Cannot be live-out if there are no successors.
     return !MBB->succ_empty();
   }
@@ -603,7 +527,7 @@ bool RegAllocFastImpl::mayLiveOut(Register VirtReg) {
     // Find the first def in the self loop MBB.
     for (const MachineInstr &DefInst : MRI->def_instructions(VirtReg)) {
       if (DefInst.getParent() != MBB) {
-        MayLiveAcrossBlocks.set(Register::virtReg2Index(VirtReg));
+        MayLiveAcrossBlocks.set(VirtReg.virtRegIndex());
         return true;
       } else {
         if (!SelfLoopDef || dominates(PosIndexes, DefInst, *SelfLoopDef))
@@ -611,7 +535,7 @@ bool RegAllocFastImpl::mayLiveOut(Register VirtReg) {
       }
     }
     if (!SelfLoopDef) {
-      MayLiveAcrossBlocks.set(Register::virtReg2Index(VirtReg));
+      MayLiveAcrossBlocks.set(VirtReg.virtRegIndex());
       return true;
     }
   }
@@ -622,7 +546,7 @@ bool RegAllocFastImpl::mayLiveOut(Register VirtReg) {
   unsigned C = 0;
   for (const MachineInstr &UseInst : MRI->use_nodbg_instructions(VirtReg)) {
     if (UseInst.getParent() != MBB || ++C >= Limit) {
-      MayLiveAcrossBlocks.set(Register::virtReg2Index(VirtReg));
+      MayLiveAcrossBlocks.set(VirtReg.virtRegIndex());
       // Cannot be live-out if there are no successors.
       return !MBB->succ_empty();
     }
@@ -632,7 +556,7 @@ bool RegAllocFastImpl::mayLiveOut(Register VirtReg) {
       // value inside a self looping block.
       if (SelfLoopDef == &UseInst ||
           !dominates(PosIndexes, *SelfLoopDef, UseInst)) {
-        MayLiveAcrossBlocks.set(Register::virtReg2Index(VirtReg));
+        MayLiveAcrossBlocks.set(VirtReg.virtRegIndex());
         return true;
       }
     }
@@ -643,7 +567,7 @@ bool RegAllocFastImpl::mayLiveOut(Register VirtReg) {
 
 /// Returns false if \p VirtReg is known to not be live into the current block.
 bool RegAllocFastImpl::mayLiveIn(Register VirtReg) {
-  if (MayLiveAcrossBlocks.test(Register::virtReg2Index(VirtReg)))
+  if (MayLiveAcrossBlocks.test(VirtReg.virtRegIndex()))
     return !MBB->pred_empty();
 
   // See if the first \p Limit def of the register are all in the current block.
@@ -651,7 +575,7 @@ bool RegAllocFastImpl::mayLiveIn(Register VirtReg) {
   unsigned C = 0;
   for (const MachineInstr &DefInst : MRI->def_instructions(VirtReg)) {
     if (DefInst.getParent() != MBB || ++C >= Limit) {
-      MayLiveAcrossBlocks.set(Register::virtReg2Index(VirtReg));
+      MayLiveAcrossBlocks.set(VirtReg.virtRegIndex());
       return !MBB->pred_empty();
     }
   }
@@ -719,17 +643,6 @@ void RegAllocFastImpl::spill(MachineBasicBlock::iterator Before,
   // pointing to this register because they are all pointing to spilled value
   // now.
   LRIDbgOperands.clear();
-
-  SmallVectorImpl<MachineInstr *> &NonStackDbgDefs =
-      VReg2NonStackDbgDefs[VirtReg];
-  for (MachineInstr *DbgDef : NonStackDbgDefs) {
-    DbgDef->getDebugReferrer().ChangeToFrameIndex(FI);
-    auto I = Lifetime2DbgKills.find(DbgDef->getDebugLifetime());
-    if (I != Lifetime2DbgKills.end())
-      I->second.handleSpill();
-  }
-  NonStackDbgDefs.clear();
-  VReg2DanglingDbgDefs[VirtReg].clear();
 }
 
 /// Insert reload instruction for \p PhysReg before \p Before.
@@ -756,8 +669,8 @@ MachineBasicBlock::iterator RegAllocFastImpl::getMBBBeginInsertionPoint(
       continue;
     }
 
-    // Most reloads should be inserted after prolog instructions.
-    if (!TII->isBasicBlockPrologue(*I))
+    // Skip prologues and inlineasm_br spills to place reloads afterwards.
+    if (!TII->isBasicBlockPrologue(*I) && !mayBeSpillFromInlineAsmBr(*I))
       break;
 
     // However if a prolog instruction reads a register that needs to be
@@ -817,7 +730,7 @@ void RegAllocFastImpl::reloadAtBegin(MachineBasicBlock &MBB) {
 /// Handle the direct use of a physical register.  Check that the register is
 /// not used by a virtreg. Kill the physreg, marking it free. This may add
 /// implicit kills to MO->getParent() and invalidate MO.
-bool RegAllocFastImpl::usePhysReg(MachineInstr &MI, MCPhysReg Reg) {
+bool RegAllocFastImpl::usePhysReg(MachineInstr &MI, MCRegister Reg) {
   assert(Register::isPhysicalRegister(Reg) && "expected physreg");
   bool displacedAny = displacePhysReg(MI, Reg);
   setPhysRegState(Reg, regPreAssigned);
@@ -825,7 +738,7 @@ bool RegAllocFastImpl::usePhysReg(MachineInstr &MI, MCPhysReg Reg) {
   return displacedAny;
 }
 
-bool RegAllocFastImpl::definePhysReg(MachineInstr &MI, MCPhysReg Reg) {
+bool RegAllocFastImpl::definePhysReg(MachineInstr &MI, MCRegister Reg) {
   bool displacedAny = displacePhysReg(MI, Reg);
   setPhysRegState(Reg, regPreAssigned);
   return displacedAny;
@@ -834,7 +747,7 @@ bool RegAllocFastImpl::definePhysReg(MachineInstr &MI, MCPhysReg Reg) {
 /// Mark PhysReg as reserved or free after spilling any virtregs. This is very
 /// similar to defineVirtReg except the physreg is reserved instead of
 /// allocated.
-bool RegAllocFastImpl::displacePhysReg(MachineInstr &MI, MCPhysReg PhysReg) {
+bool RegAllocFastImpl::displacePhysReg(MachineInstr &MI, MCRegister PhysReg) {
   bool displacedAny = false;
 
   for (MCRegUnit Unit : TRI->regunits(PhysReg)) {
@@ -844,6 +757,8 @@ bool RegAllocFastImpl::displacePhysReg(MachineInstr &MI, MCPhysReg PhysReg) {
       assert(LRI != LiveVirtRegs.end() && "datastructures in sync");
       MachineBasicBlock::iterator ReloadBefore =
           std::next((MachineBasicBlock::iterator)MI.getIterator());
+      while (mayBeSpillFromInlineAsmBr(*ReloadBefore))
+        ++ReloadBefore;
       reload(ReloadBefore, VirtReg, LRI->PhysReg);
 
       setPhysRegState(LRI->PhysReg, regFree);
@@ -863,7 +778,7 @@ bool RegAllocFastImpl::displacePhysReg(MachineInstr &MI, MCPhysReg PhysReg) {
   return displacedAny;
 }
 
-void RegAllocFastImpl::freePhysReg(MCPhysReg PhysReg) {
+void RegAllocFastImpl::freePhysReg(MCRegister PhysReg) {
   LLVM_DEBUG(dbgs() << "Freeing " << printReg(PhysReg, TRI) << ':');
 
   MCRegUnit FirstUnit = *TRI->regunits(PhysReg).begin();
@@ -911,7 +826,7 @@ unsigned RegAllocFastImpl::calcSpillCost(MCPhysReg PhysReg) const {
 
 void RegAllocFastImpl::assignDanglingDebugValues(MachineInstr &Definition,
                                                  Register VirtReg,
-                                                 MCPhysReg Reg) {
+                                                 MCRegister Reg) {
   auto UDBGValIter = DanglingDbgValues.find(VirtReg);
   if (UDBGValIter == DanglingDbgValues.end())
     return;
@@ -944,57 +859,20 @@ void RegAllocFastImpl::assignDanglingDebugValues(MachineInstr &Definition,
   Dangling.clear();
 }
 
-void RegAllocFastImpl::assignDanglingDebugDefs(MachineInstr &Definition,
-                                           Register VirtReg, MCPhysReg Reg) {
-  LLVM_DEBUG(dbgs() << "Assigning dangling DBG_DEFs for "
-                    << printReg(VirtReg, TRI) << '\n');
-  auto DanglingDbgDefsIter = VReg2DanglingDbgDefs.find(VirtReg);
-  if (DanglingDbgDefsIter == VReg2DanglingDbgDefs.end())
-    return;
-
-  SmallVectorImpl<MachineInstr *> &DanglingDbgDefs =
-      DanglingDbgDefsIter->second;
-  for (MachineInstr *DbgDef : DanglingDbgDefs) {
-    // Test whether the physreg survives from the definition to the DBG_VALUE.
-    MachineOperand &Referrer = DbgDef->getDebugReferrer();
-    bool ShouldRemove = false;
-    for (MachineBasicBlock::iterator I = std::next(Definition.getIterator()),
-                                     E = DbgDef->getIterator();
-         I != E; ++I) {
-      if (I->modifiesRegister(Reg, TRI)) {
-        LLVM_DEBUG(dbgs() << "Register did not survive for " << Referrer
-                          << '\n');
-        ShouldRemove = true;
-        break;
-      }
-    }
-    if (ShouldRemove) {
-      LLVM_DEBUG(dbgs() << "Dropping dangling DBG_DEF: " << *DbgDef << '\n');
-      removeDebugDef(*DbgDef);
-    } else {
-      Referrer.setReg(Reg);
-      Referrer.setIsRenamable();
-      killDebugDefWithinBlock(*DbgDef);
-    }
-  }
-  DanglingDbgDefs.clear();
-}
-
 /// This method updates local state so that we know that PhysReg is the
 /// proper container for VirtReg now.  The physical register must not be used
 /// for anything else when this is called.
 void RegAllocFastImpl::assignVirtToPhysReg(MachineInstr &AtMI, LiveReg &LR,
-                                           MCPhysReg PhysReg) {
+                                           MCRegister PhysReg) {
   Register VirtReg = LR.VirtReg;
   LLVM_DEBUG(dbgs() << "Assigning " << printReg(VirtReg, TRI) << " to "
                     << printReg(PhysReg, TRI) << '\n');
   assert(LR.PhysReg == 0 && "Already assigned a physreg");
   assert(PhysReg != 0 && "Trying to assign no register");
   LR.PhysReg = PhysReg;
-  setPhysRegState(PhysReg, VirtReg);
+  setPhysRegState(PhysReg, VirtReg.id());
 
   assignDanglingDebugValues(AtMI, VirtReg, PhysReg);
-  assignDanglingDebugDefs(AtMI, VirtReg, PhysReg);
 }
 
 static bool isCoalescable(const MachineInstr &MI) { return MI.isFullCopy(); }
@@ -1130,6 +1008,7 @@ void RegAllocFastImpl::allocVirtRegUndef(MachineOperand &MO) {
 
   LiveRegMap::iterator LRI = findLiveVirtReg(VirtReg);
   MCPhysReg PhysReg;
+  bool IsRenamable = true;
   if (LRI != LiveVirtRegs.end() && LRI->PhysReg) {
     PhysReg = LRI->PhysReg;
   } else {
@@ -1143,6 +1022,7 @@ void RegAllocFastImpl::allocVirtRegUndef(MachineOperand &MO) {
       // basic.
       PhysReg = getErrorAssignment(*LRI, *MO.getParent(), RC);
       LRI->Error = true;
+      IsRenamable = false;
     } else
       PhysReg = AllocationOrder.front();
   }
@@ -1153,7 +1033,7 @@ void RegAllocFastImpl::allocVirtRegUndef(MachineOperand &MO) {
     MO.setSubReg(0);
   }
   MO.setReg(PhysReg);
-  MO.setIsRenamable(true);
+  MO.setIsRenamable(IsRenamable);
 }
 
 /// Variation of defineVirtReg() with special handling for livethrough regs
@@ -1257,10 +1137,10 @@ bool RegAllocFastImpl::defineVirtReg(MachineInstr &MI, unsigned OpNum,
     LRI->Reloaded = false;
   }
   if (MI.getOpcode() == TargetOpcode::BUNDLE) {
-    BundleVirtRegsMap[VirtReg] = PhysReg;
+    BundleVirtRegsMap[VirtReg] = *LRI;
   }
   markRegUsedInInstr(PhysReg);
-  return setPhysReg(MI, MO, PhysReg);
+  return setPhysReg(MI, MO, *LRI);
 }
 
 /// Allocates a register for a VirtReg use.
@@ -1306,10 +1186,10 @@ bool RegAllocFastImpl::useVirtReg(MachineInstr &MI, MachineOperand &MO,
   LRI->LastUse = &MI;
 
   if (MI.getOpcode() == TargetOpcode::BUNDLE) {
-    BundleVirtRegsMap[VirtReg] = LRI->PhysReg;
+    BundleVirtRegsMap[VirtReg] = *LRI;
   }
   markRegUsedInInstr(LRI->PhysReg);
-  return setPhysReg(MI, MO, LRI->PhysReg);
+  return setPhysReg(MI, MO, *LRI);
 }
 
 /// Query a physical register to use as a filler in contexts where the
@@ -1321,10 +1201,9 @@ MCPhysReg RegAllocFastImpl::getErrorAssignment(const LiveReg &LR,
   MachineFunction &MF = *MI.getMF();
 
   // Avoid repeating the error every time a register is used.
-  bool EmitError = !MF.getProperties().hasProperty(
-      MachineFunctionProperties::Property::FailedRegAlloc);
+  bool EmitError = !MF.getProperties().hasFailedRegAlloc();
   if (EmitError)
-    MF.getProperties().set(MachineFunctionProperties::Property::FailedRegAlloc);
+    MF.getProperties().setFailedRegAlloc();
 
   // If the allocation order was empty, all registers in the class were
   // probably reserved. Fall back to taking the first register in the class,
@@ -1363,16 +1242,27 @@ MCPhysReg RegAllocFastImpl::getErrorAssignment(const LiveReg &LR,
 /// Changes operand OpNum in MI the refer the PhysReg, considering subregs.
 /// \return true if MI's MachineOperands were re-arranged/invalidated.
 bool RegAllocFastImpl::setPhysReg(MachineInstr &MI, MachineOperand &MO,
-                                  MCPhysReg PhysReg) {
+                                  const LiveReg &Assignment) {
+  MCPhysReg PhysReg = Assignment.PhysReg;
+  assert(PhysReg && "assignments should always be to a valid physreg");
+
+  if (LLVM_UNLIKELY(Assignment.Error)) {
+    // Make sure we don't set renamable in error scenarios, as we may have
+    // assigned to a reserved register.
+    if (MO.isUse())
+      MO.setIsUndef(true);
+  }
+
   if (!MO.getSubReg()) {
     MO.setReg(PhysReg);
-    MO.setIsRenamable(true);
+    MO.setIsRenamable(!Assignment.Error);
     return false;
   }
 
   // Handle subregister index.
-  MO.setReg(PhysReg ? TRI->getSubReg(PhysReg, MO.getSubReg()) : MCRegister());
-  MO.setIsRenamable(true);
+  MO.setReg(TRI->getSubReg(PhysReg, MO.getSubReg()));
+  MO.setIsRenamable(!Assignment.Error);
+
   // Note: We leave the subreg number around a little longer in case of defs.
   // This is so that the register freeing logic in allocateInstruction can still
   // recognize this as subregister defs. The code there will clear the number.
@@ -1832,6 +1722,8 @@ void RegAllocFastImpl::handleDebugValue(MachineInstr &MI) {
     Register Reg = MO.getReg();
     if (!Reg.isVirtual())
       continue;
+    if (!shouldAllocateRegister(Reg))
+      continue;
 
     // Already spilled to a stackslot?
     int SS = StackSlotForVirtReg[Reg];
@@ -1845,14 +1737,13 @@ void RegAllocFastImpl::handleDebugValue(MachineInstr &MI) {
     // See if this virtual register has already been allocated to a physical
     // register or spilled to a stack slot.
     LiveRegMap::iterator LRI = findLiveVirtReg(Reg);
-    SmallVector<MachineOperand *> DbgOps;
-    for (MachineOperand &Op : MI.getDebugOperandsForReg(Reg))
-      DbgOps.push_back(&Op);
+    SmallVector<MachineOperand *> DbgOps(
+        llvm::make_pointer_range(MI.getDebugOperandsForReg(Reg)));
 
     if (LRI != LiveVirtRegs.end() && LRI->PhysReg) {
       // Update every use of Reg within MI.
       for (auto &RegMO : DbgOps)
-        setPhysReg(MI, *RegMO, LRI->PhysReg);
+        setPhysReg(MI, *RegMO, *LRI);
     } else {
       DanglingDbgValues[Reg].push_back(&MI);
     }
@@ -1863,82 +1754,6 @@ void RegAllocFastImpl::handleDebugValue(MachineInstr &MI) {
   }
 }
 
-void RegAllocFastImpl::killDebugDefWithinBlock(MachineInstr &MI) {
-  DILifetime *Lifetime = MI.getDebugLifetime();
-  MCPhysReg Reg = MI.getDebugReferrer().getReg();
-  // We have no knowledge of when PhysReg might be killed
-  // outside of this block, so we insert a DbgKill before the first of the
-  // following we encounter after the DbgDef:
-  // * An existing DbgKill in the block (if it exists; we temporarily end up
-  // with two DbgKill for this case to avoid an edge case in the Live/Dead
-  // bookkeeping in DbgKills).
-  // * A clobber of PhysReg.
-  // * The first terminator of the block.
-  // Record it in Lifetime2DbgKills so it can be removed if we later spill
-  // VirtReg.
-  MachineBasicBlock::iterator InsertBefore =
-      MI.getParent()->getFirstTerminator();
-  for (MachineBasicBlock::iterator I = std::next(MI.getIterator()),
-                                   E = InsertBefore;
-       I != E; ++I) {
-    if (I->modifiesRegister(Reg, TRI) ||
-        (I->isDebugKill() && I->getDebugLifetime() == Lifetime)) {
-      InsertBefore = std::next(I);
-      break;
-    }
-  }
-  auto *DbgDefInstr =
-      BuildMI(*MBB, InsertBefore, DebugLoc(), TII->get(TargetOpcode::DBG_KILL))
-          .getInstr();
-  DbgDefInstr->addOperand(MI.getOperand(0));
-  Lifetime2DbgKills[Lifetime].handleNewDbgKill(*DbgDefInstr);
-}
-
-void RegAllocFastImpl::handleDebugDefKill(MachineInstr &MI) {
-  DILifetime *Lifetime = MI.getDebugLifetime();
-
-  if (MI.isDebugKill())
-    return Lifetime2DbgKills[Lifetime].handleExistingDbgKill(MI);
-
-  MachineOperand &Referrer = MI.getDebugReferrer();
-
-  // Ignore DBG_VALUEs that aren't based on virtual registers. These are
-  // mostly constants and frame indices.
-  if (!Referrer.isReg() || !Register::isVirtualRegister(Referrer.getReg()))
-    return;
-
-  Register VirtReg = Referrer.getReg();
-
-  // Already spilled to a stackslot?
-  int SS = StackSlotForVirtReg[VirtReg];
-  if (SS != -1) {
-    LLVM_DEBUG(dbgs() << "Rewrite DBG_DEF for spilled memory: " << MI);
-    Referrer.ChangeToFrameIndex(SS);
-    return;
-  }
-
-  // Record this DbgDef operand to update if we spill VirtReg.
-  VReg2NonStackDbgDefs[VirtReg].push_back(&MI);
-
-  LiveRegMap::iterator LRI = findLiveVirtReg(VirtReg);
-  if (LRI != LiveVirtRegs.end() && LRI->PhysReg) {
-    LLVM_DEBUG(dbgs() << "Record non-dangling DBG_DEF: " << MI);
-    // This VirtReg has a definition reaching this DbgDef, and is currently in
-    // LRI->PhysReg.
-    setPhysReg(MI, Referrer, LRI->PhysReg);
-    killDebugDefWithinBlock(MI);
-  } else {
-    LLVM_DEBUG(dbgs() << "Record dangling DBG_DEF: " << MI);
-    // There is no reaching definition of VirtReg (i.e. the DbgDef appears
-    // after the killing use of VirtReg, or there is no use), so we record it
-    // here as "dangling". After the block is allocated we will attempt to
-    // validate that the nearest definition of a PhysReg assigned to VirtReg
-    // reaches this DbgDef; if it does, we will update the DbgDef to that
-    // PhysReg, otherwise we will delete it.
-    VReg2DanglingDbgDefs[VirtReg].push_back(&MI);
-  }
-}
-
 void RegAllocFastImpl::handleBundle(MachineInstr &MI) {
   MachineBasicBlock::instr_iterator BundledMI = MI.getIterator();
   ++BundledMI;
@@ -1951,8 +1766,7 @@ void RegAllocFastImpl::handleBundle(MachineInstr &MI) {
       if (!Reg.isVirtual() || !shouldAllocateRegister(Reg))
         continue;
 
-      DenseMap<Register, MCPhysReg>::iterator DI;
-      DI = BundleVirtRegsMap.find(Reg);
+      DenseMap<Register, LiveReg>::iterator DI = BundleVirtRegsMap.find(Reg);
       assert(DI != BundleVirtRegsMap.end() && "Unassigned virtual register");
 
       setPhysReg(MI, MO, DI->second);
@@ -1986,11 +1800,6 @@ void RegAllocFastImpl::allocateBasicBlock(MachineBasicBlock &MBB) {
       continue;
     }
 
-    if (MI.isDebugDefKill()) {
-      handleDebugDefKill(MI);
-      continue;
-    }
-
     allocateInstruction(MI);
 
     // Once BUNDLE header is assigned registers, same assignments need to be
@@ -2025,11 +1834,6 @@ void RegAllocFastImpl::allocateBasicBlock(MachineBasicBlock &MBB) {
   }
   DanglingDbgValues.clear();
 
-  for (auto &I : VReg2DanglingDbgDefs)
-    for (MachineInstr *DbgDef : I.second)
-      removeDebugDef(*DbgDef);
-  VReg2DanglingDbgDefs.clear();
-
   LLVM_DEBUG(MBB.dump());
 }
 
@@ -2067,10 +1871,6 @@ bool RegAllocFastImpl::runOnMachineFunction(MachineFunction &MF) {
 
   StackSlotForVirtReg.clear();
   LiveDbgValueMap.clear();
-  VReg2NonStackDbgDefs.clear();
-  for (auto I : Lifetime2DbgKills)
-    I.second.removeDead();
-  Lifetime2DbgKills.clear();
   return true;
 }
 
