@@ -186,11 +186,13 @@
 #include "llvm/ADT/STLExtras.h"
 #include "llvm/ADT/SetOperations.h"
 #include "llvm/Analysis/CallGraph.h"
+#include "llvm/Analysis/ScopedNoAliasAA.h"
 #include "llvm/CodeGen/TargetPassConfig.h"
 #include "llvm/IR/Constants.h"
 #include "llvm/IR/DIBuilder.h"
 #include "llvm/IR/DebugInfo.h"
 #include "llvm/IR/DerivedTypes.h"
+#include "llvm/IR/Dominators.h"
 #include "llvm/IR/IRBuilder.h"
 #include "llvm/IR/InlineAsm.h"
 #include "llvm/IR/Instructions.h"
@@ -238,7 +240,7 @@ cl::opt<LoweringKind> LoweringKindLoc(
                    "Lower via mixture of above strategies")));
 
 template <typename T> std::vector<T> sortByName(std::vector<T> &&V) {
-  llvm::sort(V.begin(), V.end(), [](const auto *L, const auto *R) {
+  llvm::sort(V, [](const auto *L, const auto *R) {
     return L->getName() < R->getName();
   });
   return {std::move(V)};
@@ -287,8 +289,8 @@ class AMDGPULowerModuleLDS {
     BasicBlock *Entry = &Func->getEntryBlock();
     IRBuilder<> Builder(Entry, Entry->getFirstNonPHIIt());
 
-    Function *Decl =
-        Intrinsic::getDeclaration(Func->getParent(), Intrinsic::donothing, {});
+    Function *Decl = Intrinsic::getOrInsertDeclaration(
+        Func->getParent(), Intrinsic::donothing, {});
 
     Value *UseInstance[1] = {
         Builder.CreateConstInBoundsGEP1_32(SGV->getValueType(), SGV, 0)};
@@ -326,7 +328,7 @@ public:
     for (GlobalVariable *GV : Variables) {
       auto ConstantGepIt = LDSVarsToConstantGEP.find(GV);
       if (ConstantGepIt != LDSVarsToConstantGEP.end()) {
-        auto elt = ConstantExpr::getPtrToInt(ConstantGepIt->second, I32);
+        auto *elt = ConstantExpr::getPtrToInt(ConstantGepIt->second, I32);
         Elements.push_back(elt);
       } else {
         Elements.push_back(PoisonValue::get(I32));
@@ -531,13 +533,10 @@ public:
     // block to spare deduplicating it later.
     auto [It, Inserted] = tableKernelIndexCache.try_emplace(F);
     if (Inserted) {
-      Function *Decl =
-          Intrinsic::getDeclaration(&M, Intrinsic::amdgcn_lds_kernel_id, {});
-
       auto InsertAt = F->getEntryBlock().getFirstNonPHIOrDbgOrAlloca();
       IRBuilder<> Builder(&*InsertAt);
 
-      It->second = Builder.CreateCall(Decl, {});
+      It->second = Builder.CreateIntrinsic(Intrinsic::amdgcn_lds_kernel_id, {});
     }
 
     return It->second;
@@ -577,7 +576,7 @@ public:
 
       if (OrderedKernels.size() > UINT32_MAX) {
         // 32 bit keeps it in one SGPR. > 2**32 kernels won't fit on the GPU
-        report_fatal_error("Unimplemented LDS lowering for > 2**32 kernels");
+        reportFatalUsageError("unimplemented LDS lowering for > 2**32 kernels");
       }
 
       for (size_t i = 0; i < OrderedKernels.size(); i++) {
@@ -637,7 +636,8 @@ public:
         if (K.second.size() == 1) {
           KernelAccessVariables.insert(GV);
         } else {
-          report_fatal_error(
+          // FIXME: This should use DiagnosticInfo
+          reportFatalUsageError(
               "cannot lower LDS '" + GV->getName() +
               "' to kernel access as it is reachable from multiple kernels");
         }
@@ -649,7 +649,8 @@ public:
           ModuleScopeVariables.insert(GV);
         } else if (K.second.size() == 1) {
           KernelAccessVariables.insert(GV);
-        } else if (set_is_subset(K.second, HybridModuleRootKernels)) {
+        } else if (K.second == HybridModuleRootKernels &&
+                   set_is_subset(K.second, HybridModuleRootKernels)) {
           ModuleScopeVariables.insert(GV);
         } else {
           TableLookupVariables.insert(GV);
@@ -786,7 +787,7 @@ public:
       // backend) difficult to use. This does mean that llvm test cases need
       // to name the kernels.
       if (!Func.hasName()) {
-        report_fatal_error("Anonymous kernels cannot use LDS variables");
+        reportFatalUsageError("anonymous kernels cannot use LDS variables");
       }
 
       std::string VarName =
@@ -852,7 +853,7 @@ public:
     }
 
     assert(func->hasName()); // Checked by caller
-    auto emptyCharArray = ArrayType::get(Type::getInt8Ty(Ctx), 0);
+    auto *emptyCharArray = ArrayType::get(Type::getInt8Ty(Ctx), 0);
     GlobalVariable *N = new GlobalVariable(
         M, emptyCharArray, false, GlobalValue::ExternalLinkage, nullptr,
         Twine("llvm.amdgcn." + func->getName() + ".dynlds"), nullptr, GlobalValue::NotThreadLocal, AMDGPUAS::LOCAL_ADDRESS,
@@ -882,7 +883,7 @@ public:
         if (KernelsThatIndirectlyAllocateDynamicLDS.contains(func)) {
           assert(isKernelLDS(func));
           if (!func->hasName()) {
-            report_fatal_error("Anonymous kernels cannot use LDS variables");
+            reportFatalUsageError("anonymous kernels cannot use LDS variables");
           }
 
           GlobalVariable *N =
@@ -892,8 +893,8 @@ public:
 
           markUsedByKernel(func, N);
 
-          auto emptyCharArray = ArrayType::get(Type::getInt8Ty(Ctx), 0);
-          auto GEP = ConstantExpr::getGetElementPtr(
+          auto *emptyCharArray = ArrayType::get(Type::getInt8Ty(Ctx), 0);
+          auto *GEP = ConstantExpr::getGetElementPtr(
               emptyCharArray, N, ConstantInt::get(I32, 0), true);
           newDynamicLDS.push_back(ConstantExpr::getPtrToInt(GEP, I32));
         } else {
@@ -924,6 +925,126 @@ public:
     return KernelToCreatedDynamicLDS;
   }
 
+  static GlobalVariable *uniquifyGVPerKernel(Module &M, GlobalVariable *GV,
+                                             Function *KF) {
+    bool NeedsReplacement = false;
+    for (Use &U : GV->uses()) {
+      if (auto *I = dyn_cast<Instruction>(U.getUser())) {
+        Function *F = I->getFunction();
+        if (isKernelLDS(F) && F != KF) {
+          NeedsReplacement = true;
+          break;
+        }
+      }
+    }
+    if (!NeedsReplacement)
+      return GV;
+    // Create a new GV used only by this kernel and its function
+    GlobalVariable *NewGV = new GlobalVariable(
+        M, GV->getValueType(), GV->isConstant(), GV->getLinkage(),
+        GV->getInitializer(), GV->getName() + "." + KF->getName(), nullptr,
+        GV->getThreadLocalMode(), GV->getType()->getAddressSpace());
+    NewGV->copyAttributesFrom(GV);
+    for (Use &U : make_early_inc_range(GV->uses())) {
+      if (auto *I = dyn_cast<Instruction>(U.getUser())) {
+        Function *F = I->getFunction();
+        if (!isKernelLDS(F) || F == KF) {
+          U.getUser()->replaceUsesOfWith(GV, NewGV);
+        }
+      }
+    }
+    return NewGV;
+  }
+
+  bool lowerSpecialLDSVariables(
+      Module &M, LDSUsesInfoTy &LDSUsesInfo,
+      VariableFunctionMap &LDSToKernelsThatNeedToAccessItIndirectly) {
+    bool Changed = false;
+    const DataLayout &DL = M.getDataLayout();
+    // The 1st round: give module-absolute assignments
+    int NumAbsolutes = 0;
+    std::vector<GlobalVariable *> OrderedGVs;
+    for (auto &K : LDSToKernelsThatNeedToAccessItIndirectly) {
+      GlobalVariable *GV = K.first;
+      if (!isNamedBarrier(*GV))
+        continue;
+      // give a module-absolute assignment if it is indirectly accessed by
+      // multiple kernels. This is not precise, but we don't want to duplicate
+      // a function when it is called by multiple kernels.
+      if (LDSToKernelsThatNeedToAccessItIndirectly[GV].size() > 1) {
+        OrderedGVs.push_back(GV);
+      } else {
+        // leave it to the 2nd round, which will give a kernel-relative
+        // assignment if it is only indirectly accessed by one kernel
+        LDSUsesInfo.direct_access[*K.second.begin()].insert(GV);
+      }
+      LDSToKernelsThatNeedToAccessItIndirectly.erase(GV);
+    }
+    OrderedGVs = sortByName(std::move(OrderedGVs));
+    for (GlobalVariable *GV : OrderedGVs) {
+      unsigned BarrierScope = llvm::AMDGPU::Barrier::BARRIER_SCOPE_WORKGROUP;
+      unsigned BarId = NumAbsolutes + 1;
+      unsigned BarCnt = DL.getTypeAllocSize(GV->getValueType()) / 16;
+      NumAbsolutes += BarCnt;
+
+      // 4 bits for alignment, 5 bits for the barrier num,
+      // 3 bits for the barrier scope
+      unsigned Offset = 0x802000u | BarrierScope << 9 | BarId << 4;
+      recordLDSAbsoluteAddress(&M, GV, Offset);
+    }
+    OrderedGVs.clear();
+
+    // The 2nd round: give a kernel-relative assignment for GV that
+    // either only indirectly accessed by single kernel or only directly
+    // accessed by multiple kernels.
+    std::vector<Function *> OrderedKernels;
+    for (auto &K : LDSUsesInfo.direct_access) {
+      Function *F = K.first;
+      assert(isKernelLDS(F));
+      OrderedKernels.push_back(F);
+    }
+    OrderedKernels = sortByName(std::move(OrderedKernels));
+
+    llvm::DenseMap<Function *, uint32_t> Kernel2BarId;
+    for (Function *F : OrderedKernels) {
+      for (GlobalVariable *GV : LDSUsesInfo.direct_access[F]) {
+        if (!isNamedBarrier(*GV))
+          continue;
+
+        LDSUsesInfo.direct_access[F].erase(GV);
+        if (GV->isAbsoluteSymbolRef()) {
+          // already assigned
+          continue;
+        }
+        OrderedGVs.push_back(GV);
+      }
+      OrderedGVs = sortByName(std::move(OrderedGVs));
+      for (GlobalVariable *GV : OrderedGVs) {
+        // GV could also be used directly by other kernels. If so, we need to
+        // create a new GV used only by this kernel and its function.
+        auto NewGV = uniquifyGVPerKernel(M, GV, F);
+        Changed |= (NewGV != GV);
+        unsigned BarrierScope = llvm::AMDGPU::Barrier::BARRIER_SCOPE_WORKGROUP;
+        unsigned BarId = Kernel2BarId[F];
+        BarId += NumAbsolutes + 1;
+        unsigned BarCnt = DL.getTypeAllocSize(GV->getValueType()) / 16;
+        Kernel2BarId[F] += BarCnt;
+        unsigned Offset = 0x802000u | BarrierScope << 9 | BarId << 4;
+        recordLDSAbsoluteAddress(&M, NewGV, Offset);
+      }
+      OrderedGVs.clear();
+    }
+    // Also erase those special LDS variables from indirect_access.
+    for (auto &K : LDSUsesInfo.indirect_access) {
+      assert(isKernelLDS(K.first));
+      for (GlobalVariable *GV : K.second) {
+        if (isNamedBarrier(*GV))
+          K.second.erase(GV);
+      }
+    }
+    return Changed;
+  }
+
   bool runOnModule(Module &M) {
     CallGraph CG = CallGraph(M);
     bool Changed = superAlignLDSGlobals(M);
@@ -946,6 +1067,12 @@ public:
       }
     }
 
+    if (LDSUsesInfo.HasSpecialGVs) {
+      // Special LDS variables need special address assignment
+      Changed |= lowerSpecialLDSVariables(
+          M, LDSUsesInfo, LDSToKernelsThatNeedToAccessItIndirectly);
+    }
+
     // Partition variables accessed indirectly into the different strategies
     DenseSet<GlobalVariable *> ModuleScopeVariables;
     DenseSet<GlobalVariable *> TableLookupVariables;
@@ -1134,6 +1261,11 @@ private:
         continue;
       }
 
+      if (GV.isAbsoluteSymbolRef()) {
+        // If the variable is already allocated, don't change the alignment
+        continue;
+      }
+
       Align Alignment = AMDGPU::getAlign(DL, &GV);
       TypeSize GVSize = DL.getTypeAllocSize(GV.getValueType());
 
@@ -1191,12 +1323,6 @@ private:
 
     performOptimizedStructLayout(LayoutFields);
 
-    struct DIExprVarInfo {
-      GlobalVariable *Var;
-      uint64_t Offset;
-    };
-    DenseMap<DIFragment *, DIExprVarInfo> Fragment2VarInfo;
-
     struct DIExpressionVarInfo {
       GlobalVariable *Var;
       Metadata *DIVar;
@@ -1233,19 +1359,13 @@ private:
           CurrentOffset += Padding;
         }
 
-        if (isHeterogeneousDebug(M)) {
-          Fragment2VarInfo[FGV->getDbgDef()] =
-              DIExprVarInfo{FGV, CurrentOffset};
-        } else {
-          SmallVector<DIGlobalVariableExpression *, 1> OriginalGVEs;
-          FGV->getDebugInfo(OriginalGVEs);
-          for (const auto *OriginalGVE : OriginalGVEs) {
-            if (auto NewElementsRef =
-                    OriginalGVE->getExpression()->getNewElementsRef()) {
-              DIExpressionVarInfos.push_back({FGV,
-                                              OriginalGVE->getRawVariable(),
-                                              *NewElementsRef, CurrentOffset});
-            }
+        SmallVector<DIGlobalVariableExpression *, 1> OriginalGVEs;
+        FGV->getDebugInfo(OriginalGVEs);
+        for (const auto *OriginalGVE : OriginalGVEs) {
+          if (auto NewElementsRef =
+                  OriginalGVE->getExpression()->getNewElementsRef()) {
+            DIExpressionVarInfos.push_back({FGV, OriginalGVE->getRawVariable(),
+                                            *NewElementsRef, CurrentOffset});
           }
         }
 
@@ -1271,73 +1391,34 @@ private:
         false);
     SGV->setAlignment(StructAlign);
 
-    if (isHeterogeneousDebug(M)) {
-      DIBuilder DBuilder(M);
-
-      DIFragment *DbgVarFragment = DBuilder.createFragment();
-      SGV->setDbgDef(DbgVarFragment);
-
-      if (NamedMDNode *RN = M.getNamedMetadata("llvm.dbg.retainedNodes")) {
-        for (MDNode *O : RN->operands()) {
-          auto *L = dyn_cast<DILifetime>(O);
-          if (!L)
-            continue;
-
-          if (L->argObjects().empty())
-            continue;
-
-          // FIXME(KZHURAVL): Handle more than one arg object?
-          auto *F = dyn_cast<DIFragment>(*L->argObjectsBegin());
-          if (!F)
-            continue;
-
-          auto FragmentIterator = Fragment2VarInfo.find(F);
-          if (FragmentIterator == Fragment2VarInfo.end())
-            continue;
-
-          DIExprBuilder ExprBuilder(Ctx);
-          ExprBuilder.append<DIOp::Arg>(0, SGV->getType());
-          ExprBuilder.append<DIOp::Deref>(
-              FragmentIterator->second.Var->getValueType());
-          ExprBuilder.append<DIOp::Constant>(
-              ConstantInt::get(Type::getInt32Ty(Ctx), FragmentIterator->second.Offset));
-          ExprBuilder.append<DIOp::ByteOffset>(
-              FragmentIterator->second.Var->getValueType());
-
-          L->setLocation(ExprBuilder.intoExpr());
-          L->replaceOperandWith(2, DbgVarFragment);
+    for (auto VarInfo : DIExpressionVarInfos) {
+      DIExprBuilder ExprBuilder(Ctx);
+      for (auto Op : VarInfo.Expr) {
+        if (auto *ArgOp = std::get_if<DIOp::Arg>(&Op)) {
+          assert(ArgOp->getIndex() == 0u &&
+                 "DIOp-based DIExpression in DIGlobalVariableExpression must "
+                 "have only one argument");
+          Type *ArgTy = SGV->getType();
+          assert(isa<PointerType>(ArgTy));
+          Type *ResultTy = VarInfo.Var->getType();
+          assert(isa<PointerType>(ResultTy));
+          assert(ArgTy->getPointerAddressSpace() ==
+                 ResultTy->getPointerAddressSpace());
+          unsigned PointerSizeInBits =
+              DL.getPointerSizeInBits(ArgTy->getPointerAddressSpace());
+          auto *IntTy = IntegerType::get(Ctx, PointerSizeInBits);
+          ConstantData *C = ConstantInt::get(IntTy, VarInfo.Offset, true);
+          ExprBuilder.append<DIOp::Arg>(0u, ArgTy);
+          ExprBuilder.append<DIOp::Reinterpret>(IntTy);
+          ExprBuilder.append<DIOp::Constant>(C);
+          ExprBuilder.append<DIOp::Add>();
+          ExprBuilder.append<DIOp::Reinterpret>(ResultTy);
+        } else {
+          ExprBuilder.append(Op);
         }
       }
-    } else {
-      for (auto VarInfo : DIExpressionVarInfos) {
-        DIExprBuilder ExprBuilder(Ctx);
-        for (auto Op : VarInfo.Expr) {
-          if (auto *ArgOp = std::get_if<DIOp::Arg>(&Op)) {
-            assert(ArgOp->getIndex() == 0u &&
-                   "DIOp-based DIExpression in DIGlobalVariableExpression must "
-                   "have only one argument");
-            Type *ArgTy = SGV->getType();
-            assert(isa<PointerType>(ArgTy));
-            Type *ResultTy = VarInfo.Var->getType();
-            assert(isa<PointerType>(ResultTy));
-            assert(ArgTy->getPointerAddressSpace() ==
-                   ResultTy->getPointerAddressSpace());
-            unsigned PointerSizeInBits =
-                DL.getPointerSizeInBits(ArgTy->getPointerAddressSpace());
-            auto *IntTy = IntegerType::get(Ctx, PointerSizeInBits);
-            ConstantData *C = ConstantInt::get(IntTy, VarInfo.Offset, true);
-            ExprBuilder.append<DIOp::Arg>(0u, ArgTy);
-            ExprBuilder.append<DIOp::Reinterpret>(IntTy);
-            ExprBuilder.append<DIOp::Constant>(C);
-            ExprBuilder.append<DIOp::Add>();
-            ExprBuilder.append<DIOp::Reinterpret>(ResultTy);
-          } else {
-            ExprBuilder.append(Op);
-          }
-        }
-        SGV->addDebugInfo(DIGlobalVariableExpression::get(
-            Ctx, VarInfo.DIVar, ExprBuilder.intoExpression()));
-      }
+      SGV->addDebugInfo(DIGlobalVariableExpression::get(
+          Ctx, VarInfo.DIVar, ExprBuilder.intoExpression()));
     }
 
     DenseMap<GlobalVariable *, Constant *> Map;
@@ -1418,6 +1499,8 @@ private:
     if (!MaxDepth || (A == 1 && !AliasScope))
       return;
 
+    ScopedNoAliasAAResult ScopedNoAlias;
+
     for (User *U : Ptr->users()) {
       if (auto *I = dyn_cast<Instruction>(U)) {
         if (AliasScope && I->mayReadOrWriteMemory()) {
@@ -1427,7 +1510,34 @@ private:
           I->setMetadata(LLVMContext::MD_alias_scope, AS);
 
           MDNode *NA = I->getMetadata(LLVMContext::MD_noalias);
-          NA = (NA ? MDNode::intersect(NA, NoAlias) : NoAlias);
+
+          // Scoped aliases can originate from two different domains.
+          // First domain would be from LDS domain (created by this pass).
+          // All entries (LDS vars) into LDS struct will have same domain.
+
+          // Second domain could be existing scoped aliases that are the
+          // results of noalias params and subsequent optimizations that
+          // may alter thesse sets.
+
+          // We need to be careful how we create new alias sets, and
+          // have right scopes and domains for loads/stores of these new
+          // LDS variables. We intersect NoAlias set if alias sets belong
+          // to the same domain. This is the case if we have memcpy using
+          // LDS variables. Both src and dst of memcpy would belong to
+          // LDS struct, they donot alias.
+          // On the other hand, if one of the domains is LDS and other is
+          // existing domain prior to LDS, we need to have a union of all
+          // these aliases set to preserve existing aliasing information.
+
+          SmallPtrSet<const MDNode *, 16> ExistingDomains, LDSDomains;
+          ScopedNoAlias.collectScopedDomains(NA, ExistingDomains);
+          ScopedNoAlias.collectScopedDomains(NoAlias, LDSDomains);
+          auto Intersection = set_intersection(ExistingDomains, LDSDomains);
+          if (Intersection.empty()) {
+            NA = NA ? MDNode::concatenate(NA, NoAlias) : NoAlias;
+          } else {
+            NA = NA ? MDNode::intersect(NA, NoAlias) : NoAlias;
+          }
           I->setMetadata(LLVMContext::MD_noalias, NA);
         }
       }
@@ -1479,10 +1589,8 @@ public:
   const AMDGPUTargetMachine *TM;
   static char ID;
 
-  AMDGPULowerModuleLDSLegacy(const AMDGPUTargetMachine *TM_ = nullptr)
-      : ModulePass(ID), TM(TM_) {
-    initializeAMDGPULowerModuleLDSLegacyPass(*PassRegistry::getPassRegistry());
-  }
+  AMDGPULowerModuleLDSLegacy(const AMDGPUTargetMachine *TM = nullptr)
+      : ModulePass(ID), TM(TM) {}
 
   void getAnalysisUsage(AnalysisUsage &AU) const override {
     if (!TM)
