@@ -16,8 +16,6 @@
 #include "llvm/ObjCopy/ConfigManager.h"
 #include "llvm/ObjCopy/MachO/MachOConfig.h"
 #include "llvm/Object/Binary.h"
-#include "llvm/Object/OffloadBinary.h"
-#include "llvm/Object/OffloadBundle.h"
 #include "llvm/Option/Arg.h"
 #include "llvm/Option/ArgList.h"
 #include "llvm/Support/CRC.h"
@@ -286,11 +284,6 @@ static Expected<uint8_t> parseVisibilityType(StringRef VisType) {
   return type;
 }
 
-static void parseDumpOffloadBundle(StringRef URI) {
-  if (Error Err = object::extractOffloadBundleByURI(URI))
-    outs() << "Failed to extract from URI.";
-}
-
 namespace {
 struct TargetInfo {
   FileFormat Format;
@@ -315,6 +308,8 @@ static const StringMap<MachineInfo> TargetMap{
     // RISC-V
     {"elf32-littleriscv", {ELF::EM_RISCV, false, true}},
     {"elf64-littleriscv", {ELF::EM_RISCV, true, true}},
+    {"elf32-bigriscv", {ELF::EM_RISCV, false, false}},
+    {"elf64-bigriscv", {ELF::EM_RISCV, true, false}},
     // PowerPC
     {"elf32-powerpc", {ELF::EM_PPC, false, false}},
     {"elf32-powerpcle", {ELF::EM_PPC, false, true}},
@@ -734,15 +729,6 @@ objcopy::parseObjcopyOptions(ArrayRef<const char *> ArgsArr,
 
   SmallVector<const char *, 2> Positional;
 
-  ConfigManager ConfigMgr;
-  CommonConfig &Config = ConfigMgr.Common;
-  COFFConfig &COFFConfig = ConfigMgr.COFF;
-  ELFConfig &ELFConfig = ConfigMgr.ELF;
-  MachOConfig &MachOConfig = ConfigMgr.MachO;
-
-  if (InputArgs.hasArg(OBJCOPY_dump_offload_bundle))
-    Config.NeedPositional = false;
-
   for (auto *Arg : InputArgs.filtered(OBJCOPY_UNKNOWN))
     return createStringError(errc::invalid_argument, "unknown argument '%s'",
                              Arg->getAsString(InputArgs).c_str());
@@ -750,28 +736,27 @@ objcopy::parseObjcopyOptions(ArrayRef<const char *> ArgsArr,
   for (auto *Arg : InputArgs.filtered(OBJCOPY_INPUT))
     Positional.push_back(Arg->getValue());
 
-  if (Positional.empty() && Config.NeedPositional)
+  if (Positional.empty())
     return createStringError(errc::invalid_argument, "no input file specified");
 
-  if (Positional.size() > 2 && Config.NeedPositional)
+  if (Positional.size() > 2)
     return createStringError(errc::invalid_argument,
                              "too many positional arguments");
 
-  if (Arg *A = InputArgs.getLastArg(OBJCOPY_dump_offload_bundle)) {
-    for (StringRef URIStr : llvm::split(A->getValue(), ","))
-      parseDumpOffloadBundle(URIStr);
-  }
+  ConfigManager ConfigMgr;
+  CommonConfig &Config = ConfigMgr.Common;
+  COFFConfig &COFFConfig = ConfigMgr.COFF;
+  ELFConfig &ELFConfig = ConfigMgr.ELF;
+  MachOConfig &MachOConfig = ConfigMgr.MachO;
+  Config.InputFilename = Positional[0];
+  Config.OutputFilename = Positional[Positional.size() == 1 ? 0 : 1];
+  if (InputArgs.hasArg(OBJCOPY_target) &&
+      (InputArgs.hasArg(OBJCOPY_input_target) ||
+       InputArgs.hasArg(OBJCOPY_output_target)))
+    return createStringError(
+        errc::invalid_argument,
+        "--target cannot be used with --input-target or --output-target");
 
-  if (Config.NeedPositional) {
-    Config.InputFilename = Positional[0];
-    Config.OutputFilename = Positional[Positional.size() == 1 ? 0 : 1];
-    if (InputArgs.hasArg(OBJCOPY_target) &&
-        (InputArgs.hasArg(OBJCOPY_input_target) ||
-         InputArgs.hasArg(OBJCOPY_output_target)))
-      return createStringError(
-          errc::invalid_argument,
-          "--target cannot be used with --input-target or --output-target");
-  }
   if (InputArgs.hasArg(OBJCOPY_regex) && InputArgs.hasArg(OBJCOPY_wildcard))
     return createStringError(errc::invalid_argument,
                              "--regex and --wildcard are incompatible");
@@ -1097,6 +1082,14 @@ objcopy::parseObjcopyOptions(ArrayRef<const char *> ArgsArr,
           "bad format for --dump-section, expected section=file");
     Config.DumpSection.push_back(Value);
   }
+  for (auto *Arg : InputArgs.filtered(OBJCOPY_extract_section)) {
+    StringRef Value(Arg->getValue());
+    if (Value.split('=').second.empty())
+      return createStringError(
+          errc::invalid_argument,
+          "bad format for --extract-section, expected section=file");
+    Config.ExtractSection.push_back(Value);
+  }
   Config.StripAll = InputArgs.hasArg(OBJCOPY_strip_all);
   Config.StripAllGNU = InputArgs.hasArg(OBJCOPY_strip_all_gnu);
   Config.StripDebug = InputArgs.hasArg(OBJCOPY_strip_debug);
@@ -1434,26 +1427,25 @@ objcopy::parseInstallNameToolOptions(ArrayRef<const char *> ArgsArr) {
                              Arg->getAsString(InputArgs).c_str());
   for (auto *Arg : InputArgs.filtered(INSTALL_NAME_TOOL_INPUT))
     Positional.push_back(Arg->getValue());
-  if (Positional.empty() && Config.NeedPositional)
+  if (Positional.empty())
     return createStringError(errc::invalid_argument, "no input file specified");
-  if (Positional.size() > 1 && Config.NeedPositional)
+  if (Positional.size() > 1)
     return createStringError(
         errc::invalid_argument,
         "llvm-install-name-tool expects a single input file");
-  if (Config.NeedPositional) {
-    Config.InputFilename = Positional[0];
-    Config.OutputFilename = Positional[0];
+  Config.InputFilename = Positional[0];
+  Config.OutputFilename = Positional[0];
+
+  Expected<OwningBinary<Binary>> BinaryOrErr =
+      createBinary(Config.InputFilename);
+  if (!BinaryOrErr)
+    return createFileError(Config.InputFilename, BinaryOrErr.takeError());
+  auto *Binary = (*BinaryOrErr).getBinary();
+  if (!Binary->isMachO() && !Binary->isMachOUniversalBinary())
+    return createStringError(errc::invalid_argument,
+                             "input file: %s is not a Mach-O file",
+                             Config.InputFilename.str().c_str());
 
-    Expected<OwningBinary<Binary>> BinaryOrErr =
-        createBinary(Config.InputFilename);
-    if (!BinaryOrErr)
-      return createFileError(Config.InputFilename, BinaryOrErr.takeError());
-    auto *Binary = (*BinaryOrErr).getBinary();
-    if (!Binary->isMachO() && !Binary->isMachOUniversalBinary())
-      return createStringError(errc::invalid_argument,
-                               "input file: %s is not a Mach-O file",
-                               Config.InputFilename.str().c_str());
-  }
   DC.CopyConfigs.push_back(std::move(ConfigMgr));
   return std::move(DC);
 }
@@ -1492,16 +1484,13 @@ objcopy::parseBitcodeStripOptions(ArrayRef<const char *> ArgsArr,
                              Arg->getAsString(InputArgs).c_str());
 
   SmallVector<StringRef, 2> Positional;
-  if (Config.NeedPositional) {
-    for (auto *Arg : InputArgs.filtered(BITCODE_STRIP_INPUT))
-      Positional.push_back(Arg->getValue());
-    if (Positional.size() > 1)
-      return createStringError(
-          errc::invalid_argument,
-          "llvm-bitcode-strip expects a single input file");
-    assert(!Positional.empty());
-    Config.InputFilename = Positional[0];
-  }
+  for (auto *Arg : InputArgs.filtered(BITCODE_STRIP_INPUT))
+    Positional.push_back(Arg->getValue());
+  if (Positional.size() > 1)
+    return createStringError(errc::invalid_argument,
+                             "llvm-bitcode-strip expects a single input file");
+  assert(!Positional.empty());
+  Config.InputFilename = Positional[0];
 
   if (!InputArgs.hasArg(BITCODE_STRIP_output)) {
     return createStringError(errc::invalid_argument,
@@ -1563,31 +1552,27 @@ objcopy::parseStripOptions(ArrayRef<const char *> RawArgsArr,
     exit(0);
   }
 
+  SmallVector<StringRef, 2> Positional;
+  for (auto *Arg : InputArgs.filtered(STRIP_UNKNOWN))
+    return createStringError(errc::invalid_argument, "unknown argument '%s'",
+                             Arg->getAsString(InputArgs).c_str());
+  for (auto *Arg : InputArgs.filtered(STRIP_INPUT))
+    Positional.push_back(Arg->getValue());
+  std::copy(DashDash, RawArgsArr.end(), std::back_inserter(Positional));
+
+  if (Positional.empty())
+    return createStringError(errc::invalid_argument, "no input file specified");
+
+  if (Positional.size() > 1 && InputArgs.hasArg(STRIP_output))
+    return createStringError(
+        errc::invalid_argument,
+        "multiple input files cannot be used in combination with -o");
+
   ConfigManager ConfigMgr;
   CommonConfig &Config = ConfigMgr.Common;
   ELFConfig &ELFConfig = ConfigMgr.ELF;
   MachOConfig &MachOConfig = ConfigMgr.MachO;
 
-  SmallVector<StringRef, 2> Positional;
-  if (Config.NeedPositional) {
-    for (auto *Arg : InputArgs.filtered(STRIP_UNKNOWN))
-      return createStringError(errc::invalid_argument, "unknown argument '%s'",
-                               Arg->getAsString(InputArgs).c_str());
-    for (auto *Arg : InputArgs.filtered(STRIP_INPUT))
-      Positional.push_back(Arg->getValue());
-    std::copy(DashDash, RawArgsArr.end(), std::back_inserter(Positional));
-
-    if (Positional.empty())
-      return createStringError(errc::invalid_argument,
-                               "no input file specified");
-
-    if (Positional.size() > 1 && InputArgs.hasArg(STRIP_output)) {
-      return createStringError(
-          errc::invalid_argument,
-          "multiple input files cannot be used in combination with -o");
-    }
-  }
-
   if (InputArgs.hasArg(STRIP_regex) && InputArgs.hasArg(STRIP_wildcard))
     return createStringError(errc::invalid_argument,
                              "--regex and --wildcard are incompatible");
